/* **************************************************************************
 * Copyright (c) 1993, BSDI.  All rights reserved.
 * Library Name...  Object Engine V2.0
 * File Name......  objeng.cpp
 * Comments.......  This file contains the definitions for the core of the
 *                  object engine.
 *
 * ***************************************************************************/

#include "objeng.hpp"
#include <stdio.h>
#include <assert.h>
#include <limits.h>

extern ObjEng *objeng;

//-----------------------------------------------------------
ObjEng::ObjEng(char *tableDirectory, char *userName, char *netNamePath, int netCode)
{
#ifdef _Windows
    char netType[4];
    itoa(netCode, netType, 10);

    // NOTES: PXSetDefaults is essentially useless if you really want to set values.  It
    // NEVER sets values - it just gives you a warning if your values are higher than
    // those in the win.ini [Paradox Engine] section.  This code changes the values found
    // there (or puts them in!) if your values are higher than the win.ini values.
    // Note that these are hard coded rather than passed in.  In all of the apps I have
    // written these values, once settled upon, never change.  Thus I have taken them out
    // of the parameter list - change them here and compile.
    int iMaxTables = GetProfileInt("Paradox Engine", "MaxTables", -1);
    if ((iMaxTables == -1) || (iMaxTables < 12))
    {
        WriteProfileString("Paradox Engine", "MaxTables", "12");
        WriteProfileString("Paradox Engine", "RecBufs",   "24");
        WriteProfileString("Paradox Engine", "MaxFiles",  "24");
        WriteProfileString("Paradox Engine", "SwapSize",  "48");
    }

    // These values are set IF the win.ini vars have NEVER been set before.
    // If you want to MAKE SURE that something like ShareLocal is YES or NO, then
    // place the corresponding line in the next section.
    int iMaxLocks = GetProfileInt("Paradox Engine", "MaxLocks", -1);
    if (iMaxLocks == -1)
    {
        WriteProfileString("Paradox Engine", "MaxLocks", "32");
        WriteProfileString("Paradox Engine", "ShareLocal",  "NO");
        WriteProfileString("Paradox Engine", "PX35Locking", "NO");
    }
    // These are set ALWAYS.  Feel free to move these up into the previous section.
    WriteProfileString("Paradox Engine", "UserName", userName);
    WriteProfileString("Paradox Engine", "NetNamePath", netNamePath);
    WriteProfileString("Paradox Engine", "NetType", netType);

    // NOW initialize the engine...
    if ((status = PXWinInit(userName, PXSHARED)) != PXSUCCESS) return;
#else
    // Note that these are hard coded rather than passed in.  In all of the apps I have
    // written these values, once settled upon, never change.  Thus I have taken them out
    // of the parameter list - change them here and compile.
    PXSetDefaults(16, 4, 8, PXDEFAULT, 8, PXDEFAULT);
    if (netCode == NOSHARE)
	{
        if ((status = PXInit()) != PXSUCCESS) return;
	}
    else
	{
        if ((status = PXNetInit(netNamePath, netCode, userName)) != PXSUCCESS) return;
	}
#endif
    OESaveChanges = TRUE;
    strcpy(filedir, tableDirectory); 
    int i = strlen(filedir);   //Assuming usually WON'T have backslash
    if (filedir[i - 1] != '\\') 
    {
        filedir[i] = '\\';
        filedir[i + 1] = 0;
    }
    for (i = 0; i < OE_MAXTABLES; i++)
    {
        memset(&RegisteredTables[i], 0, sizeof(OEHandle));
    }
    TableInit();
}


//-----------------------------------------------------------
ObjEng::ObjEng()  //Uses all defaults.
{
#ifdef _Windows
    if ((status = PXWinInit("ObjEng", PXEXCLUSIVE)) != PXSUCCESS) return;
#else
    if ((status = PXInit()) != PXSUCCESS) return;
#endif
    OESaveChanges = TRUE;
    strcpy(filedir, ".\\");
    for (int i = 0; i < OE_MAXTABLES; i++)
    {
        memset(&RegisteredTables[i], 0, sizeof(OEHandle));
    }
    TableInit();
}


//-----------------------------------------------------------
ObjEng::~ObjEng()
{
    PXTblClose(tHdl);
    // Must close all of the tHdls in the RegisteredTables list.
    for (int i = 0; i < OE_MAXTABLES; i++)
    {
        if (RegisteredTables[i].tHdl != 0) 
            PXTblClose(RegisteredTables[i].tHdl);
        else
            break;
    }
    PXExit();
}


//-----------------------------------------------------------
int ObjEng::TableInit()
{
    char dbname[TBLDIRBUFLEN + 12];
    int bExist;

    strcpy(dbname, filedir);
    strcat(dbname, "objeng");  //User name of objeng should cause error.
    PXTblExist( dbname, &bExist );
    if ( ! bExist ) 
    {
        char *fldnames[] = {"ID"};
        char *types[] = {"N"};
        if ((status = PXTblCreate( dbname, 1, fldnames, types )) != PXSUCCESS) return status;
        if ((status = PXTblOpen( dbname, &tHdl, 0, 1)) != PXSUCCESS) return status;
        if ((status = PXRecBufOpen(tHdl, &rHdl)) != PXSUCCESS) return status;
        PXPutLong(rHdl, 1, 10L);
        PXRecInsert(tHdl, rHdl);
        if ((status = PXRecGoto(tHdl, 1)) != PXSUCCESS) return status;
    }
    else
    {
        if ((status = PXTblOpen( dbname, &tHdl, 0, 1)) != PXSUCCESS) return status;
        if ((status = PXRecBufOpen(tHdl, & rHdl)) != PXSUCCESS) return status;
        if ((status = PXRecGoto(tHdl, 1)) != PXSUCCESS) return status;
    }
    return status;
}


//-----------------------------------------------------------
// Maintain object counter to keep update/insert honest...
int ObjEng::GetIndex(long &index)
{
    if ((status = PXRecGet(tHdl, rHdl)) != PXSUCCESS) return -1;
    PXGetLong(rHdl, 1, &index);
    PXPutLong(rHdl, 1, index+1);
    status = PXRecUpdate(tHdl, rHdl);
    return(status);
}


//-----------------------------------------------------------
int ObjEng::EnsureExist(char *tblname, int nFields, char **FieldNames, char **FieldTypes)
{ 
    char tblPath[TBLDIRBUFLEN + 12];
    int bExist;

    // Must be given a table name and at least one field.
    assert(strlen(tblname) > 0);
    assert(nFields > 0);

    // Create path to table
    strcpy(tblPath, filedir);
    strcat(tblPath, tblname);
    PXTblExist( tblPath, &bExist );

    if ( bExist == FALSE )
        return(status = PXTblCreate( tblPath, nFields, FieldNames, FieldTypes));
    else
        return(-1);  // Success- but files were not CREATED

}


//-----------------------------------------------------------
int ObjEng::RegisterCompKey(char *tblname, int nIdxFields, FIELDHANDLE *IdxFieldHandles, char *keyName, int idxType, int caseMode)
{ 
    char tblPath[TBLDIRBUFLEN + 12];
    FIELDHANDLE compHdl[2];

    // MUST be one of these 2 constants in YOUR call to this fn./ nIdxFields is tightly constrained
    assert(idxType == SECONDARY || idxType == INCSECONDARY);
    assert(nIdxFields > 0 && nIdxFields < 17);
    assert(strlen(tblname) > 0);

    // Create path to table
    strcpy(tblPath, filedir);
    strcat(tblPath, tblname);

    // Create the intermediate handle and then use it to create the key.  Note that YOU will
    // use the secondary index by the NAME you gave it in the Code Generator!
    if ((status = PXKeyMap(tblPath, nIdxFields, IdxFieldHandles, keyName, caseMode, &(compHdl[0]))) != PXSUCCESS) return status;
    return(status = PXKeyAdd(tblPath, 1, compHdl, idxType));
}


//-----------------------------------------------------------
int ObjEng::RegisterKey(char *tblname, int nIdxFields, FIELDHANDLE *IdxFieldHandles, int idxType)
{ 
    char tblPath[TBLDIRBUFLEN + 12];

    assert(idxType == PRIMARY || idxType == SECONDARY || idxType == INCSECONDARY);
    assert(strlen(tblname) > 0);
    assert(nIdxFields > 0 && nIdxFields < 17);

    strcpy(tblPath, filedir);
    strcat(tblPath, tblname);

    return(status = PXKeyAdd( tblPath, nIdxFields, IdxFieldHandles, idxType ));
}


//-----------------------------------------------------------
int ObjEng::DropKey(char *tblname, FIELDHANDLE fldID)
{ 
    char tblPath[TBLDIRBUFLEN + 12];

    assert(strlen(tblname) > 0);
    strcpy(tblPath, filedir);
    strcat(tblPath, tblname);

    return(status = PXKeyDrop( tblPath, fldID ));
}


//-----------------------------------------------------------
int ObjEng::Register(OEHandle &oehandle)
{ 
    char tblPath[TBLDIRBUFLEN + 12];
    int i;

    // Rename using the file directory stored in dbcc
    strcpy(tblPath, filedir);
    strcat(tblPath, oehandle.TableName);

    // Move to first open slot...
    for (i = 0; ((i < OE_MAXTABLES) && (RegisteredTables[i].tHdl != 0)); i++)
    if (i == OE_MAXTABLES) return(PXERR_OUTOFTABLEHANDLES);
    assert(RegisteredTables[i].tHdl == 0);
    assert(strlen(RegisteredTables[i].TableName) == 0);

    // Put table name and index into the table.
    strcpy(RegisteredTables[i].TableName, oehandle.TableName);
    RegisteredTables[i].indexID = oehandle.indexID;

    if ((status = PXTblOpen(tblPath, &(RegisteredTables[i].tHdl), oehandle.indexID, OESaveChanges)) != PXSUCCESS) return status;
    oehandle.tHdl = RegisteredTables[i].tHdl;

    if ((status = PXRecBufOpen(RegisteredTables[i].tHdl, &(RegisteredTables[i].rHdl))) != PXSUCCESS) return status;
    oehandle.rHdl = RegisteredTables[i].rHdl;

    return(PXSUCCESS);
}

 
//-----------------------------------------------------------
int ObjEng::IsRegistered(OEHandle &oehandle)
{ 
    for (int i = 0; (i < OE_MAXTABLES) && (RegisteredTables[i].tHdl != 0); i++)
    {
        if ((strcmp(RegisteredTables[i].TableName, oehandle.TableName) == 0) &&
            (RegisteredTables[i].indexID == oehandle.indexID))
        {
            oehandle.tHdl = RegisteredTables[i].tHdl;
            oehandle.rHdl = RegisteredTables[i].rHdl;
            return TRUE;
        }
    }
    return FALSE;
}

 
// $$$$$$$$$$$$$$$$$$$ Persistant Class Definitions $$$$$$$$$$$$$$$$$$$

//-----------------------------------------------------------
PersistClass::PersistClass(char *name, int indexID)
{
    objectID = 0;
    status = PXSUCCESS;
    memset(&oehandle, 0, sizeof(OEHandle));
    strcpy(oehandle.TableName, name);
    oehandle.indexID = indexID;
}


//-------------------------------------------------------
void PersistClass::Copy(PersistClass &pc)
{
    objectID = pc.GetobjectID();
    status = pc.status;
    memcpy(&oehandle, &pc.oehandle, sizeof(OEHandle));
}


//-------------------------------------------------------
void PersistClass::Unlink()
{
    objectID = 0;
    status = PXSUCCESS;
}


//-------------------------------------------------------
int PersistClass::SynchronizeDB()
{
    FIELDHANDLE fld;

    if ((status = PXFldHandle(oehandle.tHdl, "objectID", &fld)) != PXSUCCESS) return status;
    if (objectID == 0) // Not yet added, give Index, then add
    {
        if ((status = objeng->GetIndex(objectID)) != PXSUCCESS) return status;
        PXPutLong(oehandle.rHdl, fld, objectID);
        status = PXRecInsert(oehandle.tHdl, oehandle.rHdl);
    }
    else  // 'Tis an update.
    {
        PXPutLong(oehandle.rHdl, fld, objectID);
        status = PXSrchFld(oehandle.tHdl, oehandle.rHdl, fld, SEARCHFIRST);
        if (status == PXSUCCESS)
            status = PXRecUpdate(oehandle.tHdl, oehandle.rHdl);
        else
            status = PXRecInsert(oehandle.tHdl, oehandle.rHdl);
    }
    return(status);
}


//-------------------------------------------------------
inline int PersistClass::NRecs(RECORDNUMBER & nrecs)
{
    return(status = PXTblNRecs(oehandle.tHdl, & nrecs));
}


//-------------------------------------------------------
int PersistClass::RecNum(RECORDNUMBER & recnum)
{
    return(status = PXRecNum(oehandle.tHdl, & recnum));
}


//-----------------------------------------------------------
inline int PersistClass::LinkToKey(int, int)
{
    return PXERR_TABLENOTINDEXED;
}


//-----------------------------------------------------------
inline int PersistClass::LinkToSKey(char *, int)
{
    return PXERR_TABLENOTINDEXED;
}


//-------------------------------------------------------
int PersistClass::LinkToRecord(int position)
{
    RECORDNUMBER nrecs;

    NRecs(nrecs);
    if ((position > nrecs) || (position < 1))
        return(status = PXERR_OUTOFRANGE);

    if ((status = PXRecGoto(oehandle.tHdl, position)) == PXSUCCESS)
        return Retrieve();
    else 
        return status;
}


//-----------------------------------------------------------
int PersistClass::LinkToField(char *fieldname, void *value, int mode)
{
    FIELDHANDLE fld;

    if ((status = PXFldHandle(oehandle.tHdl, fieldname, &fld)) != PXSUCCESS) return status;

    switch (localtypes[fld - 1])  // handles start at 1, C arrays at 0
    {
        case 'D' : 
            OEPutDate(fld, (char *)value);
            break;
        case 'S' :
            PXPutShort(oehandle.rHdl, fld, *(short *)value);
            break;
        case 'L' :
            PXPutLong(oehandle.rHdl, fld, *(long *)value);
            break;
        case '$' :
            PXPutDoub(oehandle.rHdl, fld, double(*(float *)value));
            break;
        case 'N' :
            PXPutDoub(oehandle.rHdl, fld, *(double *)value);
            break;
        case 'A' :
            PXPutAlpha(oehandle.rHdl, fld, (char *)value);
            break;
        default:
            status = PXERR_INVFIELDHANDLE;
            return(status);
    }
    status = PXSrchFld(oehandle.tHdl, oehandle.rHdl, fld, mode);             
    if ((status == PXSUCCESS) || ((mode == CLOSESTRECORD) && (status == PXERR_RECNOTFOUND)))
        return Retrieve();
    else 
        return status;
}


//-----------------------------------------------------------
int PersistClass::LinkToID(const long &ID)
{
    FIELDHANDLE fld;

    if (ID == 0) // ID 0 is illegal
    {
        status = PXERR_RECNOTFOUND;
        return status;
    }

	if (ID != -1)  // -1 means "Refresh", it will cause the object to Re-link using current oID
	{
		objectID = ID;
	}
    if ((status = PXFldHandle(oehandle.tHdl, "objectID", &fld)) != PXSUCCESS) return status;
    PXPutLong(oehandle.rHdl, fld, objectID);
    if (fld == 1)
	{
        if ((status = PXSrchKey(oehandle.tHdl, oehandle.rHdl, 1, SEARCHFIRST)) != PXSUCCESS) return status;
	}
    else
	{
        if ((status = PXSrchFld(oehandle.tHdl, oehandle.rHdl, fld, SEARCHFIRST)) != PXSUCCESS) return status;
	}
    return Retrieve();
}

 
//-------------------------------------------------------
int PersistClass::LinkToFirst()
{
    if ((status = PXRecFirst(oehandle.tHdl)) == PXSUCCESS)
        return Retrieve();
    else
        return status;
} 

 
//-------------------------------------------------------
int PersistClass::LinkToLast()
{
    if ((status = PXRecLast(oehandle.tHdl)) == PXSUCCESS)
        return Retrieve();
    else
        return status;
} 


//-------------------------------------------------------
int PersistClass::LinkToNext()
{
    if ((status = PXRecNext(oehandle.tHdl)) == PXSUCCESS)
        return Retrieve();
    else
        return status;
}


//-------------------------------------------------------
int PersistClass::LinkToPrev()
{
    if ((status = PXRecPrev(oehandle.tHdl)) == PXSUCCESS)
        return Retrieve();
    else
        return status;
}

 
//-----------------------------------------------------------
int PersistClass::Destroy()
{
    // Cannot use global nFields! (It changes for each derived class!)
    FIELDHANDLE fld;

    if (objectID == 0) // Not yet added->can't be deleted.
    {
        status = PXERR_RECNOTFOUND;
        return status;
    }

    if ((status = PXFldHandle(oehandle.tHdl, "objectID", &fld)) != PXSUCCESS) return status;
    PXPutLong(oehandle.rHdl, fld, objectID);
    if (fld == 1)
	{
        if ((status = PXSrchKey(oehandle.tHdl, oehandle.rHdl, 1, SEARCHFIRST)) != PXSUCCESS) return status;
	}
    else
	{
        if ((status = PXSrchFld(oehandle.tHdl, oehandle.rHdl, fld, SEARCHFIRST)) != PXSUCCESS) return status;
	}
    objectID = 0;
    return(status = PXRecDelete(oehandle.tHdl));
}

 
//-----------------------------------------------------------
int PersistClass::LockFile(int lockType)
{
    return(status = PXNetFileLock(oehandle.TableName, lockType));
}

 
//-----------------------------------------------------------
int PersistClass::UnlockFile(int lockType)
{
    return(status = PXNetFileUnlock(oehandle.TableName, lockType));
}

 
//-----------------------------------------------------------
int PersistClass::LockRecord(LOCKHANDLE &lckHandle)
{
    if (SynchronizeDB() == PXSUCCESS)
        return(status = PXNetRecLock(oehandle.tHdl, &lckHandle));
    else
        return(status);
}

 
//-----------------------------------------------------------
int PersistClass::IsRecordLocked(int &isLocked)
{

    isLocked = FALSE;
    status = PXSUCCESS;
    if (objectID == 0) return(status);
    else  // Move to this record and then ask.
    {
	    FIELDHANDLE fld;

	    if ((status = PXFldHandle(oehandle.tHdl, "objectID", &fld)) != PXSUCCESS) return status;
    	PXPutLong(oehandle.rHdl, fld, objectID);
	    if (fld == 1)
		{
    	    if ((status = PXSrchKey(oehandle.tHdl, oehandle.rHdl, 1, SEARCHFIRST)) != PXSUCCESS) return status;
		}
    	else
		{
	        if ((status = PXSrchFld(oehandle.tHdl, oehandle.rHdl, fld, SEARCHFIRST)) != PXSUCCESS) return status;
		}

        if (status == PXSUCCESS)
            return(status = PXNetRecLocked(oehandle.tHdl, &isLocked));
        else
            return(status);
    }
}


//-----------------------------------------------------------
int PersistClass::UnlockRecord(int lckHandle)
{
    return(status = PXNetRecUnlock(oehandle.tHdl, lckHandle));
}

 
//-----------------------------------------------------------
int PersistClass::GotoLock(LOCKHANDLE lckHandle)
{
    if ((status = PXNetRecGotoLock(oehandle.tHdl, lckHandle)) == PXSUCCESS)
        return Retrieve();
    else 
        return status;
}

 
//-----------------------------------------------------------
int PersistClass::LockTable(int lockType)
{
    return(status = PXNetTblLock(oehandle.tHdl, lockType));
}

 
//-----------------------------------------------------------
int PersistClass::UnlockTable(int lockType)
{
    return(status = PXNetTblUnlock(oehandle.tHdl, lockType));
}

 
//-----------------------------------------------------------
int PersistClass::IsTblChanged(int &isChanged)
{
    return(status = PXNetTblChanged(oehandle.tHdl, &isChanged));
}


//-----------------------------------------------------------
int PersistClass::TableRefresh()
{
    return(status = PXNetTblRefresh(oehandle.tHdl));
}

#ifdef _ZAPPZAPP_INCLUDED
//-------------------------------------------------------
int PersistClass::OEGetZString(int fld, int fldsize, zString &inString)
{
    static char temp[80];

	assert(fldsize < 80);
    status = PXGetAlpha(oehandle.rHdl, fld, fldsize, temp);
    if (status == PXSUCCESS) 
        inString = temp;
    else 
        inString = "";
    return(status);
}


//-------------------------------------------------------
int PersistClass::OEGetDate(const int &fld, zString &indate)
{
    long dbdate;
    int mon, day, year;
    char temp[11];
 
    if ((status = PXGetDate(oehandle.rHdl, fld, & dbdate)) == PXSUCCESS)
    {
        PXDateDecode(dbdate, &mon, &day, &year);
        if (year > 100) year %= 100;
        sprintf(temp, "%2d/%2d/%2d", mon, day, year);
        for (mon = 0; mon < 8; mon++)
            if (temp[mon] == ' ') temp[mon] = '0';
        indate = temp;
    }
    return status;
}
#endif

//-------------------------------------------------------
int PersistClass::OEPutDate(const int &fld, const char *indate)
{
    long dbdate;
    int mon, day, year;

    sscanf(indate, "%d/%d/%d", &mon, &day, &year);
    PXDateEncode(mon, day, year, &dbdate);
    return (PXPutDate(oehandle.rHdl, fld, dbdate));
}


//-------------------------------------------------------
int PersistClass::OEGetDate(const int &fld, char *indate)
{
    long dbdate;
    int mon, day, year;
 
    if ((status = PXGetDate(oehandle.rHdl, fld, & dbdate)) == PXSUCCESS)
    {
        PXDateDecode(dbdate, &mon, &day, &year);
        if (year > 100) year %= 100;
        sprintf(indate, "%2d/%2d/%2d", mon, day, year);
        for (mon = 0; mon < 8; mon++)
            if (indate[mon] == ' ') indate[mon] = '0';
    }
    return status;
}


//-------------------------------------------------------
int PersistClass::OEPutDate(const int &fld, const struct tm & indate)
{
    long dbdate;

    PXDateEncode(indate.tm_mon + 1, indate.tm_mday, indate.tm_year, & dbdate);
    return (PXPutDate(oehandle.rHdl, fld, dbdate));
}


//-------------------------------------------------------
int PersistClass::OEGetDate(const int &fld, struct tm & indate)
{
    long dbdate;
 
    if ((status = PXGetDate(oehandle.rHdl, fld, & dbdate)) == PXSUCCESS)
    {
        PXDateDecode(dbdate, & indate.tm_mon, & indate.tm_mday, & indate.tm_year);
        indate.tm_mon--;
    }
    return status;
}


// ******************************************************************
// $$$$$$$$$$$$$$$$$$$$$ Blob Class Definitions $$$$$$$$$$$$$$$$$$$$$
// ******************************************************************
Blob::Blob(OEHandle &oeh, FIELDHANDLE fld)
{
    rHdl = oeh.rHdl;
    fHdl = fld;
    status = PXSUCCESS;
}


//-----------------------------------------------------------
Blob &Blob::operator=(const Blob &inblobobj)
{
    rHdl = inblobobj.GetRHdl();
    fHdl = inblobobj.GetFHdl();
    status = inblobobj.GetStatus();
    return(*this);
}


//-----------------------------------------------------------
int Blob::OEPut(char *buff, unsigned size)
{
    if ((status = PXBlobOpenWrite(rHdl, fHdl, &bHdl, size, PXBLOBNEW)) != PXSUCCESS) return status;
    if ((status = PXBlobPut(bHdl, size, 0, (void *)buff)) != PXSUCCESS) return status;
    return (status = PXBlobClose(bHdl, PXBLOBACCEPT));
}


//-----------------------------------------------------------
int Blob::OEGet(BlobData &bdata)
{
    unsigned long dwBytes;

    // Get the handle to the blob
    if ((status = PXBlobOpenRead(rHdl, fHdl, &bHdl)) != PXSUCCESS) return status;
    // Get the size and allocate the buffer
    PXBlobGetSize(bHdl, &dwBytes);
    if ((dwBytes == 0) || (dwBytes > UINT_MAX)) 
    {
        PXBlobClose(bHdl, PXBLOBREJECT);
        return(PXERR_BLOBINVSIZE);
    }
    bdata.size = (unsigned)dwBytes;
    bdata.data = new char[bdata.size];
    // Fill the buffer using the BlobGet Function.
    if ((status = PXBlobGet(bHdl, bdata.size, 0, bdata.data)) != PXSUCCESS) return status;
    return(PXBlobClose(bHdl, PXBLOBACCEPT));
}


